"""Vectorized one-dimensional convolution."""

from functools import partial

import jax
import jax.numpy as jnp
import numpy as np

from .utils import _get_idxs, _get_slicing

_convolve_vec = jax.vmap(partial(jnp.convolve, mode="full"), (1, None), 1)
_convolve_mat = jax.vmap(_convolve_vec, (None, 1), -1)


@jax.jit
def _reshape_convolve_2d_kernel(tensor, kernel):
    """
    Reshape and convolve a N-Dimensional tensor with a 2D kernel.

    Parameters
    ----------
    tensor : jax.numpy.ndarray
        The input tensor to be convolved.
    kernel : jax.numpy.ndarray
        The convolution kernel.

    Returns
    -------
    jax.numpy.ndarray:
        The convolved tensor.
    """
    out = _convolve_mat(tensor.reshape(tensor.shape[0], -1), kernel)
    return out.reshape(
        (tensor.shape[0] + kernel.shape[0] - 1,) + tensor.shape[1:] + (kernel.shape[1],)
    )


@jax.jit
def _reshape_convolve_1d_kernel(tensor, kernel):
    """
    Reshape and convolve a N-Dimensional tensor with a 1D kernel.

    Parameters
    ----------
    tensor : jax.numpy.ndarray
        The input tensor to be convolved.
    kernel : jax.numpy.ndarray
        The convolution kernel.

    Returns
    -------
    jax.numpy.ndarray:
        The convolved tensor.
    """
    out = _convolve_vec(tensor.reshape(tensor.shape[0], -1), kernel)
    return out.reshape((tensor.shape[0] + kernel.shape[0] - 1,) + tensor.shape[1:])


@jax.jit
def _jit_tree_convolve_2d_kernel(tree, kernel):
    """
    Convolve each epoch of a tree with a 2D kernel.

    Parameters
    ----------
    tree : Any
        Tree of tensors (usually list of arrays, one array per epoch)
        to be convolved.
    kernel : jax.numpy.ndarray
        The convolution kernel.

    Returns
    -------
    jax.numpy.ndarray:
        The concatenated convolved epochs.
    """
    # Convolve each epoch
    func = partial(_reshape_convolve_2d_kernel, kernel=kernel)
    convolved_epochs = jax.tree.map(lambda x: func(x), tree)
    # Concatenate leaves on the first axis and return
    return jnp.concatenate(jax.tree.leaves(convolved_epochs), axis=0)


@jax.jit
def _jit_tree_convolve_1d_kernel(tree, kernel):
    """
    Convolve each epoch of a tree with a 1D kernel.

    Parameters
    ----------
    tree : Any
        Tree of tensors (usually list of arrays, one array per epoch) to
        be convolved.
    kernel : jax.numpy.ndarray
        The convolution kernel.

    Returns
    -------
    jax.numpy.ndarray:
        The concatenated convolved epochs.
    """
    # Convolve each epoch
    func = partial(_reshape_convolve_1d_kernel, kernel=kernel)
    convolved_epochs = jax.tree.map(lambda x: func(x), tree)
    # Concatenate leaves on the first axis and return
    return jnp.concatenate(jax.tree.leaves(convolved_epochs), axis=0)


@jax.jit
def convolve_epoch(data, kernel):
    """Convolve a single continuous temporal epoch.

    Parameters
    ----------
    data : pynapple.Tsd, pynapple.TsdFrame, pynapple.TsdTensor
        Pynapple timeseries object with data to be convolved
    kernel : numpy.ndarray, jax.numpy.ndarray
        1-D or 2-D array with kernel(s) to be used for convolution.
        First dimension is assumed to be time.

    Returns
    -------
    pynapple.Tsd, pynapple.TsdFrame, pynapple.TsdTensor
        Pynapple timeseries object with convolved data. If kernel
        is a 1-D array, the dimensions of the input data are retained.
        If kernel is a 2-D array,  another (last) dimension is added to
        store convolution with every column of kernels.
    """
    if kernel.ndim == 1:
        data = _jit_tree_convolve_1d_kernel(data, kernel)
    else:
        data = _jit_tree_convolve_2d_kernel(data, kernel)
    return data


def _get_trim_idx(t, k, trim="both"):
    if trim == "both":
        cut = ((k - 1) // 2, t + k - 1 - ((k - 1) // 2) - (1 - k % 2))
    elif trim == "left":
        cut = (k - 1, t + k - 1)
    elif trim == "right":
        cut = (0, t)
    return cut


def convolve_intervals(time_array, data_array, starts, ends, kernel, trim="both"):
    """Convolve over the first dimension.

    Convolve over the first dimension, vectorizing on every dimension of data,
    with every column of kernel.

    Parameters
    ----------
    data : pynapple.Tsd, pynapple.TsdFrame, pynapple.TsdTensor
        Pynapple timeseries object with data to be convolved
    kernel : numpy.ndarray, jax.numpy.ndarray
        1-D or 2-D array with kernel(s) to be used for convolution.
        First dimension is assumed to be time.

    Returns
    -------
    : jax.ndarray
        Pynapple timeseries object with convolved data. If kernel is a
        1-D array, the dimensions of the input data are retained. If kernel
        is a 2-D array, another (last) dimension is added to store
        convolution with every column of kernels.
    """
    idx_start, idx_end = _get_idxs(time_array, starts, ends)
    extra = _get_trim_idx(0, kernel.shape[0], trim)
    extra = (extra[0], extra[1] + 1)

    n = len(starts)
    idx_start_shift = idx_start + np.arange(1, n + 1) * extra[0] + np.arange(0, n) * extra[1]
    idx_end_shift = idx_end + np.arange(1, n + 1) * extra[0] + np.arange(0, n) * extra[1]

    idx = _get_slicing(idx_start_shift, idx_end_shift)

    tree = [data_array[start:end] for start, end in zip(idx_start, idx_end)]

    if kernel.ndim == 1:
        convolved_data = _jit_tree_convolve_1d_kernel(tree, kernel)
    else:
        convolved_data = _jit_tree_convolve_2d_kernel(tree, kernel)

    return convolved_data[idx]


def convolve(time_array, data_array, starts, ends, kernel, trim="both"):
    """One-dimensional convolution."""
    # Perform convolution
    if kernel.ndim == 0:
        raise IOError(
            "Provide a kernel with at least 1 dimension, current kernel has 0 dimensions"
        )

    if len(starts) == 1 and len(ends) == 1:
        cut = _get_trim_idx(data_array.shape[0], kernel.shape[0], trim)
        out = convolve_epoch(data_array, kernel)[cut[0] : cut[1]]
    else:
        out = convolve_intervals(time_array, data_array, starts, ends, kernel, trim)

    return out



import pytest
import pynapple as nap
from jax import config

config.update("jax_enable_x64", True)
nap.nap_config.set_backend("jax")

# def pytest_sessionstart(session):
#     """
#     Allows plugins and conftest files to perform initial configuration.
#     This hook is called for every plugin and initial conftest
#     file after command line options have been parsed.
#     """    
#     import pynapple as nap
#     from jax import config
#     config.update("jax_enable_x64", True)
#     nap.nap_config.set_backend("jax")
#     return






import pynapple as nap
import numpy as np
import pytest
from numpy import ufunc as _ufunc
import numpy.core.umath as _umath
import jax


ufuncs = {k:obj for k, obj in _umath.__dict__.items() if isinstance(obj, _ufunc)}

# nap.nap_config.set_backend("jax")


def test_backend():
    assert nap.nap_config.backend == "jax"

@pytest.mark.parametrize(
    "tsd",
    [
        nap.Tsd(t=np.arange(100), d=np.random.rand(100), time_units="s"),
        nap.TsdFrame(t=np.arange(100), d=np.random.rand(100, 5), time_units="s", columns=['a', 'b', 'c', 'd', 'e']),
        nap.TsdTensor(t=np.arange(100), d=np.random.rand(100, 5, 3), time_units="s")
    ],
)
class Test_Time_Series_1:

    def test_ufuncs(self, tsd):
        a = []
        for ufunc in ufuncs.values():
            print(ufunc)
            # Bit-twiddling functions
            if ufunc.__name__ in ['bitwise_and', 'bitwise_or', 'bitwise_xor',
                'invert', 'left_shift', 'right_shift', 'isnat', 'gcd', 'lcm',
                'ldexp', 'arccosh']:
                a.append(ufunc.__name__)
                pass

            elif ufunc.__name__ in ['matmul', 'dot']:
                break
                if tsd.ndim > 1:
                    x = np.random.rand(*tsd.shape[1:]).T
                    out = ufunc(tsd, x)
                    assert isinstance(out, tsd.__class__)
                    assert np.allclose(out.index, tsd.index)
                    assert np.allclose(out.values, ufunc(tsd.values, x))

                with pytest.raises(TypeError):
                    ufunc(tsd, tsd)

                a.append(ufunc.__name__)

            elif ufunc.__name__ in ['logaddexp', 'logaddexp2', 'true_divide']:
                x = np.random.rand(*tsd.shape)
                out = ufunc(tsd, x)
                assert isinstance(out, tsd.__class__)
                assert np.allclose(out.index, tsd.index)
                assert np.allclose(out.values, ufunc(tsd.values, x))

                with pytest.raises(TypeError):
                    ufunc(tsd, tsd)

                a.append(ufunc.__name__)

            elif ufunc.nin == 1 and ufunc.nout == 1:
                out = ufunc(tsd)
                assert isinstance(out, tsd.__class__)
                assert np.allclose(out.index, tsd.index)
                assert np.allclose(out.values, ufunc(tsd.values))
                a.append(ufunc.__name__)

            elif ufunc.nin == 2 and ufunc.nout == 1:
                # Testing with single number
                out = ufunc(tsd, 1)
                assert isinstance(out, tsd.__class__)
                assert np.allclose(out.index, tsd.index)
                assert np.allclose(out.values, ufunc(tsd.values, 1))

                # Testing with array
                x = np.random.rand(*tsd.shape)
                out = ufunc(tsd, x)
                assert isinstance(out, tsd.__class__)
                assert np.allclose(out.index, tsd.index)
                assert np.allclose(out.values, ufunc(tsd.values, x))

                # Raising an error with two tsd
                with pytest.raises(TypeError):
                    ufunc(tsd, tsd)

                a.append(ufunc.__name__)

            elif ufunc.nin == 3 and ufunc.nout == 1:
                # Testing with two number
                out = ufunc(tsd, 0.2, 0.6)
                assert isinstance(out, tsd.__class__)
                assert np.allclose(out.index, tsd.index)
                assert np.allclose(out.values, ufunc(tsd.values, 0.2, 0.6))

                a.append(ufunc.__name__)

    def test_funcs(self, tsd):
        # a = np.asarray(tsd)
        # assert np.allclose(a, tsd.values)

        if tsd.ndim>1:
            a = np.reshape(tsd, (np.prod(tsd.shape), 1))
            assert isinstance(a, jax.Array)

        if tsd.nap_class == "TsdTensor":
            a = np.reshape(tsd, (tsd.shape[0], np.prod(tsd.shape[1:])))
            assert isinstance(a, nap.TsdFrame)
            assert np.allclose(a.index, tsd.index)
            assert np.allclose(a.values, np.reshape(tsd.values, (tsd.shape[0], np.prod(tsd.shape[1:]))))

        a = np.ravel(tsd.values)
        assert np.allclose(a, np.ravel(tsd))

        a = np.transpose(tsd.values)
        assert np.allclose(a, np.transpose(tsd))

        a = np.expand_dims(tsd, axis=-1)
        assert a.ndim == tsd.ndim + 1
        if a.ndim == 2:
            assert isinstance(a, nap.TsdFrame)
        else:
            assert isinstance(a, nap.TsdTensor)

        a = np.expand_dims(tsd, axis=0)
        assert isinstance(a, np.ndarray)

        # if tsd.nap_class == "TsdFrame":
        #     a = np.column_stack((tsd[:,0],tsd[:,1]))
        #     assert isinstance(a, nap.TsdFrame)
        #     assert np.allclose(a.values, tsd.values[:,0:2])

        a = np.isnan(tsd)
        assert isinstance(a, tsd.__class__)
        assert np.allclose(a.values, np.isnan(tsd.values))

    def test_attributes(self, tsd):
        assert tsd.min() == tsd.values.min()

        with pytest.raises(AttributeError) as e_info:
            tsd.blabla()

        assert str(e_info.value) == "Time series object does not have the attribute blabla"

    def test_split(self, tsd):
        a = np.split(tsd, 4)
        b = np.split(tsd.values, 4)
        c = np.split(tsd.index, 4)
        for i in range(4):
            assert np.allclose(a[i].values, b[i])
            assert np.allclose(a[i].index, c[i])

        if tsd.ndim>1:
            a = np.split(tsd, 1, 1)
            assert isinstance(a, list)

        a = np.array_split(tsd, 4)
        b = np.array_split(tsd.values, 4)
        c = np.array_split(tsd.index, 4)
        for i in range(4):
            assert np.allclose(a[i].values, b[i])
            assert np.allclose(a[i].index, c[i])
        if tsd.ndim>1:
            a = np.array_split(tsd, 1, 1)
            assert isinstance(a, list)

        if tsd.ndim>1:
            a = np.vsplit(tsd, 4)
            b = np.vsplit(tsd.values, 4)
            c = np.split(tsd.index, 4)
            for i in range(4):
                assert np.allclose(a[i].values, b[i])
                assert np.allclose(a[i].index, c[i])

        if tsd.ndim == 3:
            a = np.dsplit(tsd, 1)
            b = np.dsplit(tsd.values, 1)
            c = np.split(tsd.index, 1)
            for i in range(1):
                assert np.allclose(a[i].values, b[i])
                assert np.allclose(a[i].index, c[i])

        if tsd.ndim == 2:
            a = np.hsplit(tsd, 1)
            b = np.hsplit(tsd.values, 1)
            for i in range(1):
                assert np.allclose(a[i].values, b[i])
                assert np.allclose(a[i].index, tsd.index)

    def test_operators(self, tsd):
        v = tsd.values

        a = tsd + 0.5
        assert isinstance(a, tsd.__class__)
        assert np.allclose(tsd.index, a.index)
        assert np.all(a.values == (v + 0.5))

        a = tsd - 0.5
        assert isinstance(a, tsd.__class__)
        assert np.allclose(tsd.index, a.index)
        assert np.all(a.values == (v - 0.5))

        a = tsd * 0.5
        assert isinstance(a, tsd.__class__)
        assert np.allclose(tsd.index, a.index)
        assert np.all(a.values == (v * 0.5))

        a = tsd / 0.5
        assert isinstance(a, tsd.__class__)
        assert np.allclose(tsd.index, a.index)
        assert np.all(a.values == (v / 0.5))

        a = tsd // 0.5
        assert isinstance(a, tsd.__class__)
        assert np.allclose(tsd.index, a.index)
        assert np.all(a.values == (v // 0.5))

        a = tsd % 0.5
        assert isinstance(a, tsd.__class__)
        assert np.allclose(tsd.index, a.index)
        assert np.all(a.values == (v % 0.5))

        a = tsd ** 0.5
        assert isinstance(a, tsd.__class__)
        assert np.allclose(tsd.index, a.index)
        assert np.allclose(a.values, v**0.5)

        a = tsd > 0.5
        assert isinstance(a, tsd.__class__)
        assert np.allclose(tsd.index, a.index)
        assert np.all(a.values == (v > 0.5))

        a = tsd >= 0.5
        assert isinstance(a, tsd.__class__)
        assert np.allclose(tsd.index, a.index)
        assert np.all(a.values == (v >= 0.5))

        a = tsd < 0.5
        assert isinstance(a, tsd.__class__)
        assert np.allclose(tsd.index, a.index)
        assert np.all(a.values == (v < 0.5))

        a = tsd <= 0.5
        assert isinstance(a, tsd.__class__)
        assert np.allclose(tsd.index, a.index)
        assert np.all(a.values == (v <= 0.5))

        a = tsd == 0.5
        assert isinstance(a, tsd.__class__)
        assert np.allclose(tsd.index, a.index)
        assert np.all(a.values == (v == 0.5))

        a = tsd != 0.5
        assert isinstance(a, tsd.__class__)
        assert np.allclose(tsd.index, a.index)
        assert np.all(a.values == (v != 0.5))

    def test_slice(self, tsd):
        a = tsd[0:10]
        assert isinstance(a, tsd.__class__)
        assert np.allclose(tsd.index[0:10], a.index)
        assert np.allclose(tsd.values[0:10], a.values)

        if tsd.nap_class == "TsdTensor":
            a = tsd[:,0:2,2:4]
            assert isinstance(a, nap.TsdTensor)
            assert np.allclose(tsd.index, a.index)
            assert np.allclose(tsd.values[:,0:2,2:4], a.values)

            a = tsd[:,0]
            assert isinstance(a, nap.TsdFrame)
            assert np.allclose(tsd.index, a.index)
            assert np.allclose(tsd.values[:,0], a.values)

            a = tsd[:,0,0]
            assert isinstance(a, nap.Tsd)
            assert np.allclose(tsd.index, a.index)
            assert np.allclose(tsd.values[:,0,0], a.values)

        if tsd.nap_class == "TsdFrame":
            a = tsd[:,0]
            assert isinstance(a, nap.Tsd)
            assert np.allclose(tsd.index, a.index)
            assert np.allclose(tsd.values[:,0], a.values)

            a = tsd.loc['a']
            assert isinstance(a, nap.Tsd)
            assert np.allclose(tsd.index, a.index)
            assert np.allclose(tsd.values[:,0], a.values)

            a = tsd.loc[['a', 'c']]
            assert isinstance(a, nap.TsdFrame)
            assert np.allclose(tsd.index, a.index)
            assert np.allclose(tsd.values[:,[0,2]], a.values)

    def test_sorting(self, tsd):
        with pytest.raises(TypeError):
            np.sort(tsd)

        with pytest.raises(TypeError):
            np.lexsort(tsd)

        with pytest.raises(TypeError):
            np.sort_complex(tsd)

        with pytest.raises(TypeError):
            np.partition(tsd)

        with pytest.raises(TypeError):
            np.argpartition(tsd)

    def test_searching(self, tsd):
        for func in [np.argmax, np.nanargmax, np.argmin, np.nanargmin]:

            a = func(tsd)
            assert a == func(tsd.values)

            if tsd.ndim>1:
                a = func(tsd, 1)
                if a.ndim == 1: assert isinstance(a, nap.Tsd)
                if a.ndim == 2: assert isinstance(a, nap.TsdFrame)
                assert np.allclose(a.values, func(tsd.values, 1))
                assert np.allclose(a.index, tsd.index)

        for func in [np.argwhere]:
            a = func(tsd)
            assert np.allclose(a, func(tsd.values))

        a = np.where(tsd>0.5)
        assert isinstance(a, tuple)

    def test_statistics(self, tsd):

        for func in [np.percentile, np.nanpercentile, np.quantile, np.nanquantile]:
            a = np.percentile(tsd, 50)
            assert a == np.percentile(tsd.values, 50)

            if tsd.ndim>1:
                a = np.percentile(tsd, 50, 1)
                assert isinstance(a, (nap.Tsd, nap.TsdFrame, nap.TsdTensor))

        for func in [np.median, np.average, np.mean, np.std, np.var, np.nanmedian, np.nanmean, np.nanstd, np.nanvar]:
            a = np.percentile(tsd, 50)
            assert a == np.percentile(tsd.values, 50)

            if tsd.ndim>1:
                a = np.percentile(tsd, 50, 1)
                assert isinstance(a, (nap.Tsd, nap.TsdFrame, nap.TsdTensor))

        if tsd.ndim == 2:
            a = np.corrcoef(tsd)
            assert isinstance(a, nap.TsdFrame)
            assert np.allclose(a.index, tsd.index)

            a = np.cov(tsd)
            assert isinstance(a, nap.TsdFrame)
            assert np.allclose(a.index, tsd.index)

            a = np.correlate(tsd[:,0], tsd[:,1])
            b = np.correlate(tsd[:,0].values, tsd[:,1].values)
            assert isinstance(a, np.ndarray)
            assert np.allclose(a, b)

            a = np.correlate(tsd[:,0], tsd[:,1], 'same')
            b = np.correlate(tsd[:,0].values, tsd[:,1].values, 'same')
            assert isinstance(a, nap.Tsd)
            assert np.allclose(a, b)

            a = np.correlate(tsd[:,0], tsd[:,1], 'full')
            b = np.correlate(tsd[:,0].values, tsd[:,1].values, 'full')
            assert isinstance(a, np.ndarray)
            assert np.allclose(a, b)

        a, bins = np.histogram(tsd)
        assert isinstance(a, np.ndarray)

        if tsd.ndim == 1:
            a = np.digitize(tsd, np.linspace(0, 1, 10))
            assert isinstance(a, nap.Tsd)

    def test_concatenate(self, tsd):

        with pytest.raises(RuntimeError, match=r"The order of the time series indexes should be strictly increasing and non overlapping."):
            np.concatenate((tsd, tsd), 0)

        tsd2 = tsd.__class__(t=tsd.index+150, d=tsd.values)

        a = np.concatenate((tsd, tsd2))
        assert isinstance(a, tsd.__class__)
        assert len(a) == len(tsd) + len(tsd2)
        assert np.allclose(a.values, np.concatenate((tsd.values, tsd2.values)))
        time_support = nap.IntervalSet(start=[0, 150], end=[99, 249])
        assert np.allclose(time_support.values, a.time_support.values)

        b = np.concatenate((tsd, tsd2), axis=0)
        assert np.allclose(a.values, b.values)
        assert np.allclose(a.index.values, b.index.values)

        c = np.concatenate((tsd, tsd2), 0)
        assert np.allclose(a.values, c.values)
        assert np.allclose(a.index.values, c.index.values)

        d = np.concatenate((tsd, tsd2))
        assert np.allclose(a.values, d.values)
        assert np.allclose(a.index.values, d.index.values)

        e = np.concatenate((tsd, tsd.values), 0)
        assert isinstance(e, np.ndarray)

        if tsd.ndim >= 2:
            out = np.concatenate((tsd, tsd), 1)
            assert isinstance(out, tsd.__class__)
            assert np.allclose(
                out.values,
                np.concatenate((tsd.values, tsd.values), 1)
                )

            out = np.concatenate((tsd.values, tsd), 1)
            assert isinstance(out, tsd.__class__)
            assert np.allclose(out.values,np.concatenate((tsd.values, tsd.values), 1))
            assert np.allclose(tsd.index.values,out.index.values)

            msg = 'Time indexes and time supports are not all equals up to pynapple precision. Returning numpy array!'
            with pytest.warns(match = msg):
                out = np.concatenate((tsd, tsd2), 1)
            assert isinstance(out, np.ndarray)

            iset = nap.IntervalSet(start = 0, end = 500)
            msg = "Time indexes are not all equals up to pynapple precision. Returning numpy array!"
            with pytest.warns(match=msg):
                out = np.concatenate((tsd.restrict(iset), tsd2.restrict(iset)), 1)

            msg = "Time supports are not all equals up to pynapple precision. Returning numpy array!"
            with pytest.warns(match=msg):
                out = np.concatenate((tsd, tsd.restrict(iset)), 1)

        if tsd.ndim == 3:
            out = np.concatenate((tsd, tsd), 2)
            assert isinstance(out, tsd.__class__)
            assert np.allclose(
                out.values,
                np.concatenate((tsd.values, tsd.values), 2)
                )
            out = np.concatenate((tsd.values, tsd), 2)
            assert isinstance(out, tsd.__class__)
            assert np.allclose(out.values,np.concatenate((tsd.values, tsd.values), 2))
            assert np.allclose(tsd.index.values,out.index.values)

    def test_fft(self, tsd):
        with pytest.raises(TypeError):
            np.fft.fft(tsd)



import warnings
from numbers import Number

import jax.numpy as jnp
import numpy as np
from numba import jit


def is_array_like(obj):
    """
    Check if an object is array-like.

    This function determines if an object has array-like properties.
    An object is considered array-like if it has attributes typically associated with arrays
    (such as `.shape`, `.dtype`, and `.ndim`), supports indexing, and is iterable.

    Parameters
    ----------
    obj : object
        The object to check for array-like properties.

    Returns
    -------
    bool
        True if the object is array-like, False otherwise.

    Notes
    -----
    This function uses a combination of checks for attributes (`shape`, `dtype`, `ndim`),
    indexability, and iterability to determine if the given object behaves like an array.
    It is designed to be flexible and work with various types of array-like objects, including
    but not limited to NumPy arrays and JAX arrays. However, it may not be full proof for all
    possible array-like types or objects that mimic these properties without being suitable for
    numerical operations.

    """
    # Check for array-like attributes
    has_shape = hasattr(obj, "shape")
    has_dtype = hasattr(obj, "dtype")
    has_ndim = hasattr(obj, "ndim")

    # Check for indexability (try to access the first element)
    try:
        obj[0]
        is_indexable = True
    except (TypeError, IndexError):
        is_indexable = False

    # Check for iterable property
    try:
        iter(obj)
        is_iterable = True
    except TypeError:
        is_iterable = False

    # not_tsd_type = not isinstance(obj, _AbstractTsd)

    return (
        has_shape and has_dtype and has_ndim and is_indexable and is_iterable
        # and not_tsd_type
    )


def cast_to_jax(array, array_name, suppress_conversion_warnings=False):
    """
    Convert an input array-like object to a jax Array.


    Parameters
    ----------
    array : array_like
        The input object to convert. This can be any object that `np.asarray` is capable of
        converting to a jax array, such as lists, tuples, and other array-like objects.
    array_name : str
        The name of the variable that we are converting, printed in the warning message.

    Returns
    -------
    ndarray
        A jax Array representation of the input `values`. If `values` is already a jax
        Array, it is returned unchanged. Otherwise, a new jax Array is created and returned.

    Warnings
    --------
    A warning is issued if the input `values` is not already a jax Array, indicating
    that a conversion has taken place and showing the original type of the input.

    """
    if not isinstance(array, jnp.ndarray) and not suppress_conversion_warnings:
        original_type = type(array).__name__
        warnings.warn(
            f"Converting '{array_name}' to jax.ndarray. The provided array was of type '{original_type}'.",
            UserWarning,
            stacklevel=2
        )
    return jnp.asarray(array)


def convert_to_jax_array(array, array_name, suppress_conversion_warnings=False):
    """Convert any array like object to jax Array.

    Parameters
    ----------
    array : ArrayLike

    array_name : str
        Array name if RuntimeError is raised or object is casted to numpy

    Returns
    -------
    jax.Array
        Jax array object

    Raises
    ------
    RuntimeError
        If input can't be converted to jax array
    """
    if isinstance(array, Number):
        return jnp.array([array])
    elif isinstance(array, (list, tuple)):
        return jnp.array(array)
    elif isinstance(array, jnp.ndarray):
        return array
    elif isinstance(array, np.ndarray):
        return cast_to_jax(array, array_name, suppress_conversion_warnings)
    elif is_array_like(array):
        return cast_to_jax(array, array_name, suppress_conversion_warnings)
    else:
        raise RuntimeError(
            "Unknown format for {}. Accepted formats are numpy.ndarray, list, tuple or any array-like objects.".format(
                array_name
            )
        )


@jit(nopython=True)
def _get_idxs(time_array, starts, ends):
    """
    Find indices in a sorted time array where each start and end time should be inserted.

    This function efficiently computes the insertion points for each start and end time
    into a sorted time array. The indices for start times are found such that each start time
    is inserted before the first element greater than the start time. The indices for end times
    are found such that each end time is inserted before the first element strictly greater than
    the end time.

    Parameters
    ----------
    time_array : array_like
        The sorted array of timestamps.
    starts : array_like
        An array of start times for which indices are required in `time_array`.
    ends : array_like
        An array of end times for which indices are required in `time_array`.

    Returns
    -------
    tuple of np.ndarray
        A tuple containing two arrays: the indices in `time_array` where the `starts`
        should be inserted, and the indices where the `ends` should be inserted.

    Examples
    --------
    >>> time_array = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    >>> starts = np.array([2, 3, 5])
    >>> ends = np.array([2.9, 4.9, 7])
    >>> idx_start, idx_end = _get_idxs(time_array, starts, ends)
    >>> print(idx_start, idx_end)  # prints [1 2 4] [2 4 7]
    """
    idx_start = np.searchsorted(time_array, starts)
    idx_end = np.searchsorted(time_array, ends, side="right")
    return idx_start, idx_end


def _fill_forward(
    time_target_array,
    time_array,
    data_array,
    starts,
    ends,
    out_of_range=np.nan,
):
    """
    Fill a time series forward in time with data. Useful for event trigger average
    if the data have a lower sampling rate. This function assumes that the data have
    been restricted beforehand.

    Parameters
    ----------
    time_target_array : ArrayLike
        The time to match
    time_array : ArrayLike
        The time of the data to extent
    data_array : ArrayLike
        The data to extent
    starts : ArrayLike

    ends : ArrayLike

    out_of_range : Number, optional
        How to fill the gap

    Returns
    -------
    : ArrayLike
        The data time series filled forward.

    """
    filled_d = np.full(
        (time_target_array.shape[0], *data_array.shape[1:]),
        out_of_range,
        dtype=data_array.dtype,
    )
    fill_idx = 0

    idx_start_target, idx_end_target = _get_idxs(time_target_array, starts, ends)
    idx_start, idx_end = _get_idxs(time_array, starts, ends)

    for i in range(len(idx_start)):
        d = data_array[idx_start[i] : idx_end[i]]
        t = time_array[idx_start[i] : idx_end[i]]
        ts = time_target_array[idx_start_target[i] : idx_end_target[i]]

        idxs = np.searchsorted(t, ts, side="right") - 1
        filled_d[fill_idx : fill_idx + len(ts)][idxs >= 0] = d[idxs[idxs >= 0]]
        fill_idx += len(ts)

    return filled_d


def _get_shifted_indices(idx_start, idx_end, window):
    """
    Compute shifted indices for given start and end indices based on a specified window size.

    Parameters
    ----------
    idx_start : ArrayLike
        Array of start indices.
    idx_end : ArrayLike
        Array of end indices.
    windows : ArrayLike


    Returns
    -------
    : tuple of ArrayLike
        A tuple containing two arrays: shifted start indices and shifted end indices.
    """
    cum_delta = np.concatenate([[0], np.cumsum(idx_end - idx_start)])
    idx_start_shift = window * np.arange(idx_start.shape[0]) + cum_delta[:-1]
    idx_end_shift = window * np.arange(idx_end.shape[0]) + cum_delta[1:]
    return idx_start_shift, idx_end_shift


@jit(nopython=True)
def _get_slicing(idx_start, idx_end):
    """
    Generate an array of indices for slicing, based on provided start and end indices.

    Parameters
    ----------
    idx_start : ArrayLike
        Array of start indices.
    idx_end : ArrayLike
        Array of end indices.

    Returns
    -------
    : ArrayLike
        An array of indices constructed from the start and end indices.
    """
    iend = np.sum(idx_end - idx_start)
    ix = np.zeros(iend, dtype=np.int32)
    cnt = 0
    for st, en in zip(idx_start, idx_end):
        for k in range(st, en):
            ix[cnt] = k
            cnt += 1
    return ix



"""
# threshold
"""

import numpy as np
import pynapple as nap
import jax.numpy as jnp
from time import perf_counter
import matplotlib.pyplot as plt

# %%
# Machine Configuration
import jax
print(jax.devices())



# %% 
def get_mean_perf_class(obj):
    n = 10
    tmp = np.zeros(n)
    for i in range(n):
        t1 = perf_counter()
        out = obj.threshold(0.0)
        t2 = perf_counter()
        tmp[i] = t2 - t1
    return np.mean(tmp), np.std(tmp)

def benchmark_threshold():
    times = []
    for T in np.arange(500000, 1000000, 100000):
        time_array = np.arange(T) / 2
        data_array = np.random.randn(len(time_array))
        starts = np.arange(1, T // 2 - 1, 20)
        ends = np.arange(1, T // 2 - 1, 20) + 8
        
        ep = nap.IntervalSet(start=starts, end=ends)
        tsd = nap.Tsd(t=time_array, d=data_array, time_support=ep)
        
        res = tsd.threshold(0.0) # First call to compile
        m, s = get_mean_perf_class(tsd)
        times.append([T, m, s])
    return np.array(times)

# %% 
# Calling with numba
nap.nap_config.set_backend("numba")
num_times = benchmark_threshold()


# %%
# Calling with jax

nap.nap_config.set_backend("jax")
jax_times = benchmark_threshold()


# %%
# Figure
plt.figure()
for arr, label in zip(
    [num_times, jax_times],
    ["numba backend", "pynajax backend"],
    ):
    plt.plot(arr[:, 0], arr[:, 1], "o-", label=label)
    plt.fill_between(arr[:, 0], arr[:, 1] - arr[:, 2], arr[:, 1] + arr[:, 2], alpha=0.2)
plt.legend()
plt.xlabel("Number of Time points")
plt.ylabel("Time (s)")
plt.title("Threshold benchmark")
plt.show()



import jax
import jax.numpy as jnp
import numpy as np

from pynajax.jax_core_bin_average import bin_average

from .utils import _fill_forward, _get_idxs, _get_shifted_indices, _get_slicing

# dot prod shifted counts vs 1D var y vmapped over shift
# [(n_shift, T), (T, )] -> (n_shift, )
_dot_prod = jax.vmap(lambda x, y: jnp.nansum(x * y), in_axes=(0, None), out_axes=0)
# vmap over the neurons
# [(n_shift, T, n_neurons), (T, )] -> (n_shift, n_neurons)
_dot_prod_neu = jax.vmap(_dot_prod, in_axes=(2, None), out_axes=1)
# vmap over the features
# [(n_shift, T, n_neurons), (T, n_features)] -> (n_shift, n_neurons, n_features)
_dot_prod_feature = jax.vmap(_dot_prod_neu, in_axes=(None, 1), out_axes=2)


def pad_and_roll(count_array, windows):
    """
    Pad and roll the input array to generate shifted versions of the array according
    to specified window size and padding direction.

    Parameters
    ----------
    count_array : ArrayLike
        The input array to pad and roll. This is typically a count or spike array in
        neural data analysis.
    windows : tuple of int
        The number of steps to include in the window. This defines the extent of the
        rolling operation.

    Returns
    -------
    ArrayLike
        A 2D array where each row represents the input array rolled by one step in
        the range defined by the window and padding type. Only the valid range (original
        data indices) is returned.

    Notes
    -----
    The function uses `np.nan` for padding, which may need to be considered in subsequent
    calculations. Depending on the analysis, handling of `np.nan` may be required to avoid
    statistical or computational errors.
    """
    n_samples = count_array.shape[0]
    pad = lambda x: jnp.pad(x, pad_width=(windows, (0, 0)), constant_values=np.nan)
    indices = jnp.arange(-windows[0], windows[1] + 1)[::-1]
    idx = jnp.arange(windows[0], n_samples + windows[0])
    roll = jax.vmap(lambda i: jnp.roll(pad(count_array), -i, axis=0))
    return roll(indices)[:, idx]


def event_trigger_average(
    time_target_array,
    count_array,
    time_array,
    data_array,
    starts,
    ends,
    windows,
    binsize,
    batch_size,
):
    """
    Main function to call for event-triggered averages (ETA) across multiple epochs.
    This function assumes count array and data array have been restricted before.

    Parameters
    ----------
    time_target_array : ArrayLike
        Timestamps array of counts
    count_array : ArrayLike
        Count array of events
    time_array : ArrayLike
        Timestamps of data. Can be lower or above sampling rate of time_array
    data_array : ArrayLike
        Data to average at event
    starts : ArrayLike
        Start times of the epochs over which the STA is computed.
    ends : ArrayLike
        End times of the epochs, corresponding to each start time.
    windows : tuple of int
        The number of time steps to include before and after each event for averaging.
    binsize : float
        The bin size. Used by `bin_average`.
    batch_size : int, optional
        Number of elements to process in each batch.

    Returns
    -------
    jnp.ndarray :
        The combined STA calculated across all specified epochs.

    """
    # Need to bring data_target_array to same shape as count_array
    # bin_average
    if count_array.shape[0] < data_array.shape[0]:
        time_array, data_array = bin_average(time_array, data_array, starts, ends, binsize)
    # fill_forward
    else:
        data_array = _fill_forward(time_target_array, time_array, data_array, starts, ends)

    idx_start, idx_end = _get_idxs(time_target_array, starts, ends)
    idx_start_shift, idx_end_shift = _get_shifted_indices(idx_start, idx_end, np.sum(windows) + 1)

    # get the indices for setting elements
    ix_orig = _get_slicing(idx_start, idx_end)
    ix_shift = _get_slicing(idx_start_shift, idx_end_shift)

    # define larger array
    tot_size = ix_shift[-1] - ix_shift[0] + 1

    shape = data_array.shape
    if data_array.ndim == 1:
        data_array = np.expand_dims(data_array, -1)
    else:
        data_array = data_array.reshape(shape[0], -1)

    num_full_batches = np.maximum(1, data_array.shape[1] // batch_size)
    batch_size = np.minimum(data_array.shape[1], batch_size)
    carry = 0

    tot_count = jnp.sum(count_array, 0)

    count_array = (
        jnp.full((tot_size, *count_array.shape[1:]), np.nan).at[ix_shift].set(count_array[ix_orig])
    )
    count_array = pad_and_roll(count_array, windows)

    def scan_fn(carry, x):
        slc = jax.lax.dynamic_slice(data_array, (0, carry), (data_array.shape[0], batch_size))
        slc = jnp.full((tot_size, *slc.shape[1:]), np.nan).at[ix_shift].set(slc[ix_orig])
        batch_result = _dot_prod_feature(count_array, slc)
        return carry + batch_size, batch_result

    _, res = jax.lax.scan(scan_fn, carry, None, length=num_full_batches)

    res = jnp.transpose(res, (1, 2, 0, 3))  # move features at the end
    res = res.reshape(*res.shape[:-2], -1)  # flatten features

    extra_elements = data_array.shape[1] % batch_size
    if extra_elements:
        # compute residual slice
        slc = data_array[:, -extra_elements:]
        slc = jnp.full((tot_size, *slc.shape[1:]), np.nan).at[ix_shift].set(slc[ix_orig])
        resid = _dot_prod_feature(count_array, slc)
        # resid = resid.transpose(1, 2, 0).reshape(*res.shape[:-1], -1)
        res = np.concatenate([res, resid], axis=2)

    # reshape back to original
    res = res.reshape((np.sum(windows) + 1, count_array.shape[-1], *shape[1:]))

    res = jnp.apply_along_axis(jnp.divide, 1, res, tot_count)

    return res


def perievent_continuous(data_array, N_w, N_target, slice_idx, w_starts):
    """
    Extract data from data array.

    Parameters
    ----------
    data_array : ArrayLike
        The jax array to extract the data
    N_w : int
        window size
    N_target : int
        Number of reference time points
    slice_idx : ArrayLike
        2-d numpy array of size (N_target, 2). Each row is the start and end of the slice
    w_starts : ArrayLike
        1-d numpy array of size (N_target,). Whether to trim the window on the left when slicing
    """
    new_data_array = jnp.full((N_w, N_target, *data_array.shape[1:]), jnp.nan)

    w_sizes = slice_idx[:, 1] - slice_idx[:, 0]  # Different sizes

    all_w_sizes = np.unique(w_sizes)
    all_w_start = np.unique(w_starts)

    for w_size in all_w_sizes:
        for w_start in all_w_start:
            col_idx = w_sizes == w_size
            new_idx = np.zeros((w_size, np.sum(col_idx)), dtype=int)
            for i, tmp in enumerate(slice_idx[col_idx]):
                new_idx[:, i] = np.arange(tmp[0], tmp[1])

            new_data_array = new_data_array.at[w_start : w_start + w_size, col_idx].set(
                data_array[new_idx]
            )

    return new_data_array



"""
# convolve
"""

import numpy as np
import pynapple as nap
import jax.numpy as jnp
from time import perf_counter
import matplotlib.pyplot as plt

# %%
# Machine Configuration
import jax
print(jax.devices())


# %% 
def get_mean_perf_class(obj, kernel):
    n = 10
    tmp = np.zeros(n)
    for i in range(n):
        t1 = perf_counter()
        out = obj.convolve(kernel)
        t2 = perf_counter()
        tmp[i] = t2 - t1
    return np.mean(tmp), np.std(tmp)

def benchmark_convolve(kernel):
    times = []
    for nd in range(10, 500, 50):
        print("Dimensions ", nd)
        t = np.arange(10000)
        d = np.random.randn(10000, nd)
        tsd = nap.TsdFrame(t=t, d=d)        
        tsd2 = tsd.convolve(kernel) # First call to compile
        m, s = get_mean_perf_class(tsd, kernel)
        times.append([nd, m, s])
    return np.array(times)

# %%
# Calling with jax
nap.nap_config.set_backend("jax")
jax_times = benchmark_convolve(jnp.ones(11))

# %% 
# Calling with numba
nap.nap_config.set_backend("numba")
num_times = benchmark_convolve(np.ones(11))

# %%
# Figure
plt.figure()
for arr, label in zip(
    [num_times, jax_times],
    ["numba backend", "pynajax backend"],
    ):
    plt.plot(arr[:, 0], arr[:, 1], "o-", label=label)
    plt.fill_between(arr[:, 0], arr[:, 1] - arr[:, 2], arr[:, 1] + arr[:, 2], alpha=0.2)
plt.legend()
plt.xlabel("Number of Dimensions")
plt.ylabel("Time (s)")
plt.title("Convolve benchmark")
plt.savefig("../images/convolve_benchmark.png")
plt.show()


# %% 
# Saving




import itertools

import jax.numpy as jnp
import numpy as np
import pytest

from pynajax import jax_core_convolve

from contextlib import nullcontext as does_not_raise


@pytest.mark.parametrize(
    "shape_array1, shape_array2", [((3,), (10, 2)), ((3,), (10, 2, 3))]
)
def test_2d_convolve_epoch_vec(shape_array1, shape_array2):
    """Compare convolution with numpy for 1D kernel"""
    np.random.seed(111)
    arr1 = np.random.normal(size=shape_array1)
    arr2 = np.random.normal(size=shape_array2)

    # res_numpy = []
    # for i in range(arr2.shape[1]):
    #     res_numpy.append(np.convolve(arr2[:,i], arr1, mode="full"))
    # res_numpy = np.array(res_numpy).T

    res_numpy = np.apply_along_axis(np.convolve, 0, arr2, arr1)

    res_pynajax = jax_core_convolve.convolve_epoch(jnp.asarray(arr2), arr1)
    np.testing.assert_array_almost_equal(res_pynajax, res_numpy)


@pytest.mark.parametrize(
    "shape_array1, shape_array2", [((3, 2), (10, 2)), ((3, 2), (10, 2, 3))]
)
def test_2d_convolve_epoch_mat(shape_array1, shape_array2):
    """Compare convolution with numpy for 2D kernel"""
    np.random.seed(111)
    arr1 = np.random.normal(size=shape_array1)
    arr2 = np.random.normal(size=shape_array2)

    res_numpy = []
    for i in range(arr1.shape[1]):
        res_numpy.append(np.apply_along_axis(np.convolve, 0, arr2, arr1[:,i]))
    res_numpy = np.array(res_numpy)
    res_numpy = np.rollaxis(res_numpy, 0, res_numpy.ndim)

    res_pynajax = jax_core_convolve.convolve_epoch(arr2, arr1)
    np.testing.assert_array_almost_equal(res_pynajax, res_numpy)
    assert isinstance(res_pynajax, jnp.ndarray)


@pytest.mark.parametrize(
    "ep",
    [
        np.array([[0, 100]]),
        np.array([[0, 19],[20,100]])
    ],
)
@pytest.mark.parametrize(
    "time_array",
    [
        np.arange(100),
    ],
)
@pytest.mark.parametrize(
    "data_array",
    [jnp.ones((100,)), jnp.ones((100, 1)), jnp.ones((100, 2)), jnp.ones((100, 2, 3))],
)
@pytest.mark.parametrize(
    "kernel",
    [
        np.ones((10,)),
        np.ones((10, 1)),
        np.ones((10, 2)),
        jnp.ones((10,)),
        jnp.ones((10, 1)),
        jnp.ones((10, 2)),
    ],
)
def test_convolve_intervals(time_array, data_array, ep, kernel):
    """Run convolution on single and multi interval."""
    jax_core_convolve.convolve_intervals(
        time_array, 
        data_array, 
        ep[:,0],
        ep[:,1],
        kernel
        )


@pytest.mark.parametrize(
    "iset",
    [
        np.array([[0, 100]]),
        np.array([[0, 19],[20,100]])
    ],
)
@pytest.mark.parametrize(
    "data",
    [jnp.ones((100,)), jnp.ones((100, 1)), jnp.ones((100, 2)), jnp.ones((100, 2, 3))],
)
@pytest.mark.parametrize(
    "time",
    [
        np.arange(100),
    ],
)
@pytest.mark.parametrize(
    "kernel",
    [
        np.ones((10,)),
    ],
)
def test_convolve_intervals_shape_1d_kernel(time, data, iset, kernel):
    """Check that the shape of input and output matches if kernel is 1D."""    
    expected_shape = data.shape
    res = jax_core_convolve.convolve_intervals(time, data, iset[:,0], iset[:,1], kernel)
    assert all(res.shape[i] == expected_shape[i] for i in range(data.ndim))


@pytest.mark.parametrize(
    "iset",
    [
        np.array([[0, 100]]),
        np.array([[0, 19],[20,100]])
    ],
)
@pytest.mark.parametrize(
    "data",
    [jnp.ones((100,)), jnp.ones((100, 1)), jnp.ones((100, 2)), jnp.ones((100, 2, 3))],
)
@pytest.mark.parametrize(
    "time",
    [
        np.arange(100),
    ],
)
@pytest.mark.parametrize(
    "kernel",
    [
        np.ones((10, 1)),
        np.ones((10, 2)),
    ],
)
def test_convolve_intervals_shape_2d_kernel(time, data, iset, kernel):
    """
    Check that the shape of output is equal to that of the input
    plus the second dimension of the kernel attached at the end.
    """    
    expected_shape = (*data.shape, kernel.shape[1])
    res = jax_core_convolve.convolve_intervals(time, data, iset[:,0], iset[:,1], kernel)
    assert all(res.shape[i] == expected_shape[i] for i in range(data.ndim))





@pytest.mark.parametrize(
    "ep",
    [
        np.array([[0, 100]]),
        np.array([[0, 19],[20,100]])
    ],
)
@pytest.mark.parametrize(
    "time_array",
    [
        np.arange(100),
    ],
)
@pytest.mark.parametrize(
    "data_array",
    [jnp.ones((100,)), jnp.ones((100, 1)), jnp.ones((100, 2)), jnp.ones((100, 2, 3))],
)
@pytest.mark.parametrize(
    "kernel",
    [
        np.ones((10,)),
        np.ones((10, 1)),
        np.ones((10, 2)),
        jnp.ones((10,)),
        jnp.ones((10, 1)),
        jnp.ones((10, 2)),
    ],
)
def test_convolve(time_array, data_array, ep, kernel):
    """Run convolution on single and multi interval."""
    jax_core_convolve.convolve(
        time_array, 
        data_array, 
        ep[:,0],
        ep[:,1],
        kernel
        )

def test_raise_error():
    """Run convolution on single and multi interval."""
    with pytest.raises(IOError, match=r"Provide a kernel with at least 1 dimension, current kernel has 0 dimensions"):
        jax_core_convolve.convolve(
            np.arange(100),
            jnp.ones((100,)),
            np.array([0]),
            np.array([100]),
            np.array(0)
            )


import os
import sys
import re

from mkdocs_gallery.sorting import FileNameSortKey

min_reported_time = 0
if 'SOURCE_DATE_EPOCH' in os.environ:
    min_reported_time = sys.maxint if sys.version_info[0] == 2 else sys.maxsize

# To be used as the "base" config,
# mkdocs-gallery is a port of sphinx-gallery. For a detailed list
# of configuration options see https://sphinx-gallery.github.io/stable/configuration.html
conf = {
    # report runtime if larger than this value
    'min_reported_time': min_reported_time,
    # order your section in file name alphabetical order
    'within_subsection_order': FileNameSortKey,
    # run every script that matches pattern
    # (here we match every file that ends in .py)
    'filename_pattern': re.escape(os.sep) + r"plot_.+\.py$",
    'ignore_pattern': r"_plot_.+\.py$"    
}



import os
import shutil
from pathlib import Path

import nox


def copy_and_overwrite(from_path, to_path):
    if os.path.exists(to_path):
        shutil.rmtree(to_path)
    shutil.copytree(from_path, to_path)


to_exclude = [
    "test_jitted.py",
    "test_numpy_compatibility.py",  # Pynajax has its own
    "test_config.py",
    "conftest.py",
]

# nox --no-venv -s linters
# nox --no-venv -s tests


@nox.session(name="linters")
def linters(session):
    """Run linters"""
    session.run("ruff", "check", "src", "--ignore", "D")


@nox.session(name="tests")
def tests(session):
    """Run the test suite."""

    nap_test_path = Path("../pynapple/tests")
    path = Path("tests/")

    # Copying test files
    tocopy = os.listdir(nap_test_path)
    tocopy = list(filter(lambda x: x[0:5] == "test_" and x not in to_exclude, tocopy))
    for f in tocopy:
        shutil.copy(nap_test_path.joinpath(f), path.joinpath(f))

    # Copying nwbfiletest and npzfiletest
    copy_and_overwrite(nap_test_path.joinpath("nwbfilestest"), path.joinpath("nwbfilestest"))

    try:
        session.run("pytest")  # , "--pdb", "--pdbcls=IPython.terminal.debugger:Pdb")
    finally:
        # Remove files
        for f in tocopy:
            try:
                os.remove(path.joinpath(f))
            except OSError:
                pass
        # Remove folders
        shutil.rmtree(path.joinpath("nwbfilestest"))
